#include "esp_adc_cal.h"

int ledPin = 5;
int motorPin = 25;
int pwmChannel = 0;
int analog1Pin = 0;
// NW102
// Closed - 4.5v - 472
// Idle   - 3.7v - 386
// Open   - 0.4v - 170
//
// LS2 Silver
// Closed - 4.5v - 473
// Idle   - 3.7v - 392
// Open   - 0.4v - 169
// idle-open range of 222

int analog2Pin = 2; // ranges from 037-477, 111 at idle (range of 366)   040-479, 122 at idle
// NW102
// Closed - 0.5v - 040
// Idle   - 1.4v - 122
// Open   - 4.6v - 480
//
// LS2 Silver
// Closed - 0.5v - 037
// Idle   - 1.3v - 111
// Open   - 4.6v - 477

// Current draw
//
// NW102
// Closed - 0.08
// Half   - 0.20 / 0.42
// Open   - 0.62
//
// LS2 Silver
// Closed - 0.08
// Half   - 0.2 / 0.3
// Open   - 0.4

typedef enum __State { Pause0, Open1, Pause1, Open2, Pause2, Close1, Pause3, Close2, Pause4, Open75, Open25, Open60, Open20, Close3, WOT } State;

State state = Pause0;

int duty = 40;
int sensor1Value = 0;
int sensor2Value = 0;
float acceleratorPedalPosition = 0;
float bladePosition = 0;
int sensor1Target = 0;

void ReadSensors()
{
  sensor1Value = analogRead(analog1Pin);
  sensor2Value = analogRead(analog2Pin);
}

void PrintLine()
{
    Serial.print((int)acceleratorPedalPosition);
    Serial.print(",");
    Serial.print((int)bladePosition);
    Serial.print(",");
    Serial.print(sensor1Target);
    Serial.print(",");
    Serial.print(duty);
    Serial.print(",");
    Serial.print(sensor1Value);
    Serial.print(",");
    Serial.print(sensor2Value);
    Serial.println();  
}

const int maxA = 385;
const int minA = 169;

// Given a value from 0-100, return a value from sensor1's observed range
int GetTargetSensorValue(float input)
{
  float x = 100.0 - input;
  x *= ((maxA - minA) / 100.0);
  x += minA;
  return x;
}

// Given a value from sensor1's range, return a value from 0-100
float GetBladePosition(float input)
{
  float x = input - minA;
  x /= ((maxA - minA) / 100.0);
  x = 100.0 - x;
  return x;
}

float errorAccumulator = 0;
float lastError = 0;

int CalculateDutyCycle(int target, int actual, long elapsed)
{
  float error = (actual - target);
  errorAccumulator += error;

  error *= 256;
  
  float P = error * 10;
  float I = errorAccumulator * 1.0;
  float D = (error - lastError) * 30.0;

  if (I > 20) I = 20;
  if (I < -20) I = -20;
  
  lastError = error;
  float raw = P + I + D;
  if (raw > 192 * 256) raw = 192 * 256;
  if (raw < 10) raw = 10;
  return raw;
}

void setup()
{
    // Disable motors
    pinMode(ledPin, OUTPUT);
    digitalWrite(ledPin, HIGH);
  
    // Set motor to zero
    pinMode(motorPin, OUTPUT);
    digitalWrite(motorPin, LOW);

    // Prepare for PWM control
    // Kudos to Peter at Tech Explorations for documenting this:
    // https://techexplorations.com/guides/esp32/begin/pwm/
    ledcAttachPin(motorPin, pwmChannel);
    ledcSetup(pwmChannel, 1000, 16);
    
    Serial.begin(115200);
    Serial.println("Initializing...");
    delay(1000);

    pinMode(analog1Pin, INPUT);
    pinMode(analog2Pin, INPUT);
    adcAttachPin(analog1Pin);
    adcAttachPin(analog2Pin);
    analogSetAttenuation(ADC_11db); // ADC_ATTEN_DB_11);
    analogSetWidth(ADC_WIDTH_BIT_12);
    analogSetClockDiv(255);
    
    // Enable motors
    digitalWrite(ledPin, LOW);
    Serial.println("Motors enabled...");
    delay(1000);

    // Make sure the sensor readings are stable
    for (int i = 0; i < 10; i++)
    {
      ReadSensors();
      PrintLine();
      delay(10);
    }
}

long lastTime = 0;
long pauseStart = millis();
long pauseDuration = 1000;

void loop()
{
  long currentTime = millis();
  long elapsed = currentTime - lastTime;
  long lastTime = currentTime;
  
  float delta = elapsed / (1000.0 * 1000.0);
  delta /= 10;
  
  switch (state)
  {
    case Pause0:
      acceleratorPedalPosition = 0;
      if (currentTime > pauseStart + pauseDuration)
      {
        state = Open1;
      }
      break;
    
    case Open1:
      if (acceleratorPedalPosition > 50)
      {
        pauseStart = currentTime;
        state = Pause1;
      }
      else
      {
        acceleratorPedalPosition += delta;
      }
      break;

    case Pause1:
      acceleratorPedalPosition = 50;
      if (currentTime > pauseStart + pauseDuration)
      {
        state = Open2;
      }
      break;

    case Open2:
      if (acceleratorPedalPosition > 100)
      {
        pauseStart = currentTime;
        state = Pause2;
      }
      else
      {
        acceleratorPedalPosition += delta;
      }
      break;

    case Pause2:
      acceleratorPedalPosition = 100;
      if (currentTime > pauseStart + pauseDuration)
      {
        state = Close1;
      }
      break;

    case Close1:
      if (acceleratorPedalPosition < 50)
      {
        pauseStart = currentTime;
        state = Pause3;
      }
      else
      {
        acceleratorPedalPosition -= delta;
      }
      break;

    case Pause3:
      acceleratorPedalPosition = 50;
      if (currentTime > pauseStart + pauseDuration)
      {
        state = Close2;
      }
      break;

    case Close2:
      if (acceleratorPedalPosition < 0)
      {
        pauseStart = currentTime;
        state = Pause4;
      }
      else
      {
        acceleratorPedalPosition -= delta;
      }
      break;

    case Pause4:
      acceleratorPedalPosition = 0;
      if (currentTime > pauseStart + pauseDuration)
      {
        pauseStart = currentTime;
        state = Open75;
      }
      break;

    case Open75:
      acceleratorPedalPosition = 75;
      if (currentTime > pauseStart + pauseDuration)
      {
        pauseStart = currentTime;
        state = Open25;
      }
      break;

    case Open25:
      acceleratorPedalPosition = 25;
      if (currentTime > pauseStart + pauseDuration)
      {
        pauseStart = currentTime;
        state = Open60;
      }
      break;
    
    case Open60:
      acceleratorPedalPosition = 60;
      if (currentTime > pauseStart + pauseDuration)
      {
        pauseStart = currentTime;
        state = Open20;
      }
      break;

    case Open20:
      acceleratorPedalPosition = 20;
      if (currentTime > pauseStart + pauseDuration)
      {
        pauseStart = currentTime;
        state = Close3;
      }
      break;

    case Close3:
      acceleratorPedalPosition = 0;
      if (currentTime > pauseStart + pauseDuration)
      {
        pauseStart = currentTime;
        state = WOT;
      }
      break;

    case WOT:
      acceleratorPedalPosition = 100;
      if (currentTime > pauseStart + pauseDuration)
      {
        pauseStart = currentTime;
        state = Pause0;
      }
      break;
    
    }

    ReadSensors();
    
    sensor1Target = GetTargetSensorValue(acceleratorPedalPosition);
    bladePosition = GetBladePosition(sensor1Value);

    duty = CalculateDutyCycle(sensor1Target, sensor1Value, elapsed);
    
    ledcWrite(pwmChannel, duty);

    if (currentTime % 100 == 0)
    {
      PrintLine();
    }
}
